#!/usr/bin/python
# A program running a Registry query for tables matching specific
# properties (in this case: having columns with a specific UCD)
# and then running generated queries to yield a common result
# table.
#
# In this script, we're sending the table over to TOPCAT directly.
#
# Right now (Jun 2017), RegTAP doesn't work terribly well for TAP
# table discovery yet, so we're providing a workaround here.
#
# WARNING: With this kind of thing, you can hit a lot of services
# with a lot of queries.  At least while still trying out, please
# don't run every iteration on the whole VO.
#
# In production, you probably want to avoid going through dictionaries
# for all result rows, but that's 

from astropy import table
import pyvo
import sys
import vohelper  # this should be where you got this file


# The region of interest as a circle
RA, DEC, SR = 83.8221, -5.3911, 0.01


# This is how this kind of thing should be done in the future...
REGTAP_ENDPOINT = "http://reg.g-vo.org/tap"
REGTAP_QUERY = """
SELECT TOP 1 DISTINCT access_url, table_name
FROM rr.interface
NATURAL JOIN rr.capability
NATURAL JOIN rr.res_table
NATURAL JOIN rr.table_column
NATURAL JOIN rr.stc_spatial
WHERE
	standard_id LIKE 'ivo://ivoa.net/std/tap%'
	AND ucd LIKE 'pos.pm%'
	AND 1=CONTAINS(gavo_simbadpoint('M 42'), coverage)
"""


# the data items you're interested in, expressed through UCDs
# that's pairs of (ucd, name, unit); implied is src, src_table
RESULT_SCHEMA = [
	('cat_id', "meta.id;meta.main", None),
	('ra', "pos.eq.ra;meta.main", "deg"),
	('dec', "pos.eq.dec;meta.main", "deg"),
	('pmra', "pos.pm;pos.eq.ra", "mas/yr"),
	('pmde', "pos.pm;pos.eq.dec", "mas/yr"),]


def get_services_and_tables(regtap_query):
	"""returns a sequence of (service, [table-names]) pairs for a
	hardwired RegTAP query.

	regtap_query must (at least) have access_url and table_name in
	the select clause.

	NOTE: As of mid-2017, this query doesn't really do what you'd expect
	due to deficiencies in the providers' registry records.  Use the
	GloTS variant until then.
	"""
	reg_svc = pyvo.dal.TAPService(REGTAP_ENDPOINT)
	result = reg_svc.run_sync(regtap_query)

	svcs = {}
	for row in result.table:
		svcs.setdefault(row["access_url"], []).append(row["table_name"])
	return svcs.items()


def get_services_and_tables_glots():
	"""stands in for get_services_and_tables until people get their
	act together.

	For subtle reasons, coverage queries here won't do what you want here.
	Sorry.
	"""
	reg_svc = pyvo.dal.TAPService(REGTAP_ENDPOINT)
	result = reg_svc.run_sync("""
		SELECT DISTINCT accessurl as access_url, table_name
		FROM glots.columns
			JOIN glots.services using (ivoid)
			JOIN glots.tables using (ivoid, table_name)
		WHERE 
			ucd LIKE 'pos.pm%'
			-- Skip half-broken but time-consuming wfau services for now
			AND accessurl NOT LIKE '%wfau%' 
		""", maxrec=3000)

	svcs = {}
	for row in result.table:
		svcs.setdefault(row["access_url"], []).append(row["table_name"])
	return svcs.items()


def get_query(svc, table):
	"""returns a query producing RESULT_SCHEMA for table in svc, plus
	a defaults dict to complete with empty columns.

	svc is a pyvo.dal.TAPService, table a name from the schema.
	"""
	db_table = svc.tables[table]
	select_clause, defaults = [], {}
	for dest_name, ucd, unit in RESULT_SCHEMA:
		try:
			select_clause.append("{} AS {}".format(
				vohelper.quote_if_necessary(
					vohelper.get_name_for_ucd(ucd, db_table)),
				dest_name))
		except KeyError:
			# ADQL cannot SELECT NULL, hence this workaround
			defaults[dest_name] = None

	return ("SELECT {selclause} FROM {srctable}"
		" WHERE 1=CONTAINS(POINT('', {racol}, {deccol}),"
		"   CIRCLE('', {ra}, {dec}, {sr}))").format(
			selclause=", ".join(select_clause),
			srctable=vohelper.quote_if_necessary(table),
			racol=vohelper.get_name_for_ucd("pos.eq.ra;meta.main", db_table),
			deccol=vohelper.get_name_for_ucd("pos.eq.dec;meta.main", db_table),
			ra=RA,
			dec=DEC,
			sr=SR), defaults


def get_rows_for_svc(access_url, tables):
	"""returns rows in RESULT_SCHEMA for tables in the service at access_url.
	"""
	svc = pyvo.dal.TAPService(access_url)
	for table_name in tables:
		try:
			sys.stderr.write("> {} on {}\n".format(table_name, access_url))
			query, defaults = get_query(svc, table_name)
			defaults.update({"src": access_url, "src_table": table_name})

			db_rows = svc.run_sync(query).table
			for row in db_rows:
				# TODO: Fix units here
				res = dict(zip(row.colnames, row.as_void()))
				res.update(defaults)
				yield res
		except Exception, msg:
			sys.stderr.write("Table {} at {} broken: {}\n".format(
				table_name, access_url, msg))


def make_result_table(recs):
	"""returns an astropy table for RESULT_SCHEMA plus src, src_table.
	"""
	res = table.Table()
	res.add_column(table.Column(name='src', 
		description="Source service",
		data=[r["src"] for r in recs]))
	res.add_column(table.Column(name='src_table', 
		description="Source table",
		data=[r["src_table"] for r in recs]))
	 
	for name, ucd, unit in RESULT_SCHEMA:
		res.add_column(table.Column(name=name,
			data=[r[name] for r in recs],
			meta={"ucd": ucd})),

	return res


def main():
	recs = []
	for svc_url, tables in get_services_and_tables_glots():
	# in the future: get_services_and_tables(REGTAP_ENDPOINT)
		recs.extend(
			list(get_rows_for_svc(svc_url, tables)))
	
	res_table = make_result_table(recs)
	with vohelper.SAMP_conn() as conn:
		topcat_id = vohelper.find_client(conn, "topcat")
		vohelper.send_table_to(conn, topcat_id, res_table, "lotsarows")
				

if __name__=="__main__":
	main()
